@@ -149,6 +149,25 @@ static inline void store_unlock(void)
 	mutex_unlock(&pidns_store_mutex);
 }
 
+#define define_interruptible_lock(type, lockname, lockfn)           \
+        int lockname##_interruptible(type *l)                       \
+        {                                                           \
+                int ret = ETIMEDOUT;                                \
+                while (!fuse_interrupted() && (ret == ETIMEDOUT)) { \
+                        struct timespec deadline;                   \
+                        clock_gettime(CLOCK_REALTIME, &deadline);   \
+                        deadline.tv_sec += 1;                       \
+                        ret = lockfn(l, &deadline);                 \
+                }                                                   \
+                return -ret;                                        \
+        }
+
+define_interruptible_lock(pthread_mutex_t, mutex_lock, pthread_mutex_timedlock)
+define_interruptible_lock(pthread_rwlock_t, rwlock_rdlock, pthread_rwlock_timedrdlock)
+define_interruptible_lock(pthread_rwlock_t, rwlock_wrlock, pthread_rwlock_timedwrlock)
+
+#undef define_interruptible_lock
+
 /* /proc/       =    6
  *                +
  * <pid-as-str> =   INTTYPE_TO_STRLEN(pid_t)
@@ -847,6 +866,7 @@ static void __attribute__((constructor)) lxcfs_init(void)
 {
 	__do_close int init_ns = -EBADF, root_fd = -EBADF,
 				  pidfd = -EBADF;
+	__do_free char *cgroup = NULL;
 	int i = 0;
 	pid_t pid;
 	struct hierarchy *hierarchy;
@@ -901,7 +921,8 @@ static void __attribute__((constructor)) lxcfs_init(void)
 		lxcfs_info("Kernel supports pidfds");
 	}
 
-	can_use_swap = cgroup_ops->can_use_swap(cgroup_ops);
+	cgroup = get_pid_cgroup(pid, "memory");
+	can_use_swap = cgroup && cgroup_ops->can_use_swap(cgroup_ops, cgroup);
 	if (can_use_swap)
 		lxcfs_info("Kernel supports swap accounting");
 	else
