@@ -89,13 +89,37 @@ static int start_loadavg(void)
 {
 	char *error;
 	pthread_t (*__load_daemon)(int);
+	int (*__load_daemon_v2)(pthread_t *, int);
 
+	/* try a new load_daemon_v2() API */
+	dlerror();
+	__load_daemon_v2 = (int (*)(pthread_t *, int))dlsym(dlopen_handle, "load_daemon_v2");
+	error = dlerror();
+	if (error)
+		/* try with an old symbol name */
+		goto old_api;
+
+	lxcfs_debug("start_loadavg: using load_daemon_v2");
+
+	if (__load_daemon_v2(&loadavg_pid, 1)) {
+		/* we have to NULLify loadavg_pid as in case of error it's contents are undefined */
+		loadavg_pid = 0;
+		return log_error(-1, "Failed to start loadavg daemon");
+	}
+
+	/* we are done */
+	return 0;
+
+old_api:
+	/* go with an old load_daemon() API */
 	dlerror();
 	__load_daemon = (pthread_t(*)(int))dlsym(dlopen_handle, "load_daemon");
 	error = dlerror();
 	if (error)
 		return log_error(-1, "%s - Failed to start loadavg daemon", error);
 
+	lxcfs_debug("start_loadavg: using load_daemon");
+
 	loadavg_pid = __load_daemon(1);
 	if (!loadavg_pid)
 		return -1;
@@ -637,6 +661,8 @@ static int do_sys_releasedir(const char *path, struct fuse_file_info *fi)
 	return __sys_releasedir(path, fi);
 }
 
+static bool cgroup_is_enabled = false;
+
 #if HAVE_FUSE3
 static int lxcfs_getattr(const char *path, struct stat *sb, struct fuse_file_info *fi)
 #else
@@ -657,7 +683,7 @@ static int lxcfs_getattr(const char *path, struct stat *sb)
 		return 0;
 	}
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_getattr(path, sb);
 		down_users();
@@ -688,7 +714,7 @@ static int lxcfs_opendir(const char *path, struct fuse_file_info *fi)
 	if (strcmp(path, "/") == 0)
 		return 0;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_opendir(path, fi);
 		down_users();
@@ -723,13 +749,13 @@ static int lxcfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 		    dir_filler(filler, buf, "..", 0) != 0 ||
 		    dir_filler(filler, buf, "proc", 0) != 0 ||
 		    dir_filler(filler, buf, "sys", 0) != 0 ||
-		    dir_filler(filler, buf, "cgroup", 0) != 0)
+		    (cgroup_is_enabled && dir_filler(filler, buf, "cgroup", 0) != 0))
 			return -ENOMEM;
 
 		return 0;
 	}
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_readdir(path, buf, filler, offset, fi);
 		down_users();
@@ -760,7 +786,7 @@ static int lxcfs_access(const char *path, int mode)
 	if (strcmp(path, "/") == 0 && (mode & W_OK) == 0)
 		return 0;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_access(path, mode);
 		down_users();
@@ -822,7 +848,7 @@ static int lxcfs_open(const char *path, struct fuse_file_info *fi)
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_open(path, fi);
 		down_users();
@@ -851,7 +877,7 @@ static int lxcfs_read(const char *path, char *buf, size_t size, off_t offset,
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_read(path, buf, size, offset, fi);
 		down_users();
@@ -880,7 +906,7 @@ int lxcfs_write(const char *path, const char *buf, size_t size, off_t offset,
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_write(path, buf, size, offset, fi);
 		down_users();
@@ -959,7 +985,7 @@ int lxcfs_mkdir(const char *path, mode_t mode)
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_mkdir(path, mode);
 		down_users();
@@ -977,7 +1003,7 @@ int lxcfs_chown(const char *path, uid_t uid, gid_t gid)
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_chown(path, uid, gid);
 		down_users();
@@ -1004,7 +1030,7 @@ int lxcfs_truncate(const char *path, off_t newsize, struct fuse_file_info *fi)
 int lxcfs_truncate(const char *path, off_t newsize)
 #endif
 {
-	if (strncmp(path, "/cgroup", 7) == 0)
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0)
 		return 0;
 
 	if (strncmp(path, "/sys", 4) == 0)
@@ -1017,7 +1043,7 @@ int lxcfs_rmdir(const char *path)
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_rmdir(path);
 		down_users();
@@ -1035,7 +1061,7 @@ int lxcfs_chmod(const char *path, mode_t mode)
 {
 	int ret;
 
-	if (strncmp(path, "/cgroup", 7) == 0) {
+	if (cgroup_is_enabled && strncmp(path, "/cgroup", 7) == 0) {
 		up_users();
 		ret = do_cg_chmod(path, mode);
 		down_users();
@@ -1051,6 +1077,42 @@ int lxcfs_chmod(const char *path, mode_t mode)
 	return -ENOENT;
 }
 
+#if HAVE_FUSE3
+static void fuse_intr_sighandler(int sig)
+{
+	(void) sig;
+	/* Nothing to do */
+}
+
+static int fuse_init_intr_signal(int signum)
+{
+	struct sigaction old_sa;
+	struct sigaction sa;
+
+	if (sigaction(signum, NULL, &old_sa) == -1)
+		return log_error(-1, "cannot get old signal handler\n");
+
+	if (old_sa.sa_handler != SIG_DFL)
+		return log_error(-1, "%d has non-default handler\n", signum);
+
+	memset(&sa, 0, sizeof(struct sigaction));
+
+	/*
+	 * We *must* enable SA_RESTART, otherwise we may accidentally
+	 * break some code which is not ready to signals/fuse interrupt.
+	 */
+	sa.sa_flags = SA_RESTART;
+
+	sa.sa_handler = fuse_intr_sighandler;
+	sigemptyset(&sa.sa_mask);
+
+	if (sigaction(signum, &sa, NULL) == -1)
+		return log_error(-1, "cannot set interrupt signal handler\n");
+
+	return 0;
+}
+#endif
+
 #if HAVE_FUSE3
 static void *lxcfs_init(struct fuse_conn_info *conn, struct fuse_config *cfg)
 #else
@@ -1062,6 +1124,8 @@ static void *lxcfs_init(struct fuse_conn_info *conn)
 
 #if HAVE_FUSE3
 	cfg->direct_io = 1;
+	cfg->intr = 1;
+	cfg->intr_signal = LXCFS_INTR_SIGNAL;
 #endif
 
 	return fuse_get_context()->private_data;
@@ -1119,7 +1183,7 @@ static void usage(void)
 	lxcfs_info("Options :");
 	lxcfs_info("  -d, --debug          Run lxcfs with debugging enabled");
 	lxcfs_info("  -f, --foreground     Run lxcfs in the foreground");
-	lxcfs_info("  -n, --help           Print help");
+	lxcfs_info("  -h, --help           Print help");
 	lxcfs_info("  -l, --enable-loadavg Enable loadavg virtualization");
 	lxcfs_info("  -o                   Options to pass directly through fuse");
 	lxcfs_info("  -p, --pidfile=FILE   Path to use for storing lxcfs pid");
@@ -1128,6 +1192,7 @@ static void usage(void)
 	lxcfs_info("  -v, --version        Print lxcfs version");
 	lxcfs_info("  --enable-cfs         Enable CPU virtualization via CPU shares");
 	lxcfs_info("  --enable-pidfd       Use pidfd for process tracking");
+	lxcfs_info("  --enable-cgroup      Enable cgroup emulation code");
 	exit(EXIT_FAILURE);
 }
 
@@ -1176,6 +1241,7 @@ static const struct option long_options[] = {
 
 	{"enable-cfs",		no_argument,		0,	  0	},
 	{"enable-pidfd",	no_argument,		0,	  0	},
+	{"enable-cgroup",	no_argument,		0,	  0	},
 
 	{"pidfile",		required_argument,	0,	'p'	},
 	{								},
@@ -1256,6 +1322,8 @@ int main(int argc, char *argv[])
 				opts->use_pidfd = true;
 			else if (strcmp(long_options[idx].name, "enable-cfs") == 0)
 				opts->use_cfs = true;
+			else if (strcmp(long_options[idx].name, "enable-cgroup") == 0)
+				cgroup_is_enabled = true;
 			else
 				usage();
 			break;
@@ -1396,12 +1464,21 @@ int main(int argc, char *argv[])
 	fuse_argv[fuse_argc++] = new_argv[0];
 	fuse_argv[fuse_argc] = NULL;
 
+	lxcfs_info("Starting LXCFS at %s", argv[0]);
+
 	do_reload(false);
 	if (install_signal_handler(SIGUSR1, sigusr1_reload)) {
 		lxcfs_error("%s - Failed to install SIGUSR1 signal handler", strerror(errno));
 		goto out;
 	}
 
+#if HAVE_FUSE3
+	if (fuse_init_intr_signal(LXCFS_INTR_SIGNAL)) {
+		lxcfs_error("Failed to install fuse interrupt signal handler");
+		goto out;
+	}
+#endif
+
 	if (!pidfile) {
 		snprintf(pidfile_buf, sizeof(pidfile_buf), "%s/lxcfs.pid", RUNTIME_PATH);
 		pidfile = pidfile_buf;
