@@ -82,6 +82,45 @@ static off_t get_procfile_size(const char *path)
 	return answer;
 }
 
+static off_t get_procfile_size_with_personality(const char *path)
+{
+	struct fuse_context *fc = fuse_get_context();
+	__u32 host_personality = liblxcfs_personality(), caller_personality;
+	bool change_personality;
+	int ret;
+	off_t procfile_size_ret;
+
+	if (get_task_personality(fc->pid, &caller_personality) < 0)
+		return log_error(0, "Failed to get caller process (pid: %d) personality", fc->pid);
+
+	/* do we need to change thread personality? */
+	change_personality = host_personality != caller_personality;
+
+	if (change_personality) {
+		ret = personality(caller_personality);
+		if (ret == -1)
+			return log_error(0, "Call to personality(%d) failed: %s\n",
+					caller_personality, strerror(errno));
+
+		lxcfs_debug("task (tid: %d) personality was changed %d -> %d\n",
+				(int)syscall(SYS_gettid), ret, caller_personality);
+	}
+
+	procfile_size_ret = get_procfile_size(path);
+
+	if (change_personality) {
+		ret = personality(host_personality);
+		if (ret == -1)
+			return log_error(0, "Call to personality(%d) failed: %s\n",
+					host_personality, strerror(errno));
+
+		lxcfs_debug("task (tid: %d) personality was restored %d -> %d\n",
+				(int)syscall(SYS_gettid), ret, host_personality);
+	}
+
+	return procfile_size_ret;
+}
+
 __lxcfs_fuse_ops int proc_getattr(const char *path, struct stat *sb)
 {
 	struct timespec now;
@@ -106,7 +145,10 @@ __lxcfs_fuse_ops int proc_getattr(const char *path, struct stat *sb)
 	    strcmp(path, "/proc/swaps")		== 0 ||
 	    strcmp(path, "/proc/loadavg")	== 0 ||
 	    strcmp(path, "/proc/slabinfo")	== 0) {
-		sb->st_size = get_procfile_size(path);
+		if (liblxcfs_functional())
+			sb->st_size = get_procfile_size_with_personality(path);
+		else
+			sb->st_size = get_procfile_size(path);
 		sb->st_mode = S_IFREG | 00444;
 		sb->st_nlink = 1;
 		return 0;
@@ -164,7 +206,10 @@ __lxcfs_fuse_ops int proc_open(const char *path, struct fuse_file_info *fi)
 
 	info->type = type;
 
-	info->buflen = get_procfile_size(path) + BUF_RESERVE_SIZE;
+	if (liblxcfs_functional())
+		info->buflen = get_procfile_size_with_personality(path) + BUF_RESERVE_SIZE;
+	else
+		info->buflen = get_procfile_size(path) + BUF_RESERVE_SIZE;
 
 	info->buf = zalloc(info->buflen);
 	if (!info->buf)
@@ -194,21 +239,37 @@ __lxcfs_fuse_ops int proc_release(const char *path, struct fuse_file_info *fi)
 	return 0;
 }
 
-static uint64_t get_memlimit(const char *cgroup, bool swap)
+/**
+ * Gets a non-hierarchical memory controller limit, or UINT64_MAX if no limit is
+ * in place. If `swap` is true, reads 'swap' (v2) or 'memsw' (v1); otherwise
+ * reads the memory (RAM) limits.
+ *
+ * @returns 0 on success (and sets `*limit`), < 0 on error
+ */
+static int get_memlimit(const char *cgroup, bool swap, uint64_t *limit)
 {
 	__do_free char *memlimit_str = NULL;
-	uint64_t memlimit = 0;
+	uint64_t memlimit = UINT64_MAX;
 	int ret;
 
 	if (swap)
 		ret = cgroup_ops->get_memory_swap_max(cgroup_ops, cgroup, &memlimit_str);
 	else
 		ret = cgroup_ops->get_memory_max(cgroup_ops, cgroup, &memlimit_str);
-	if (ret > 0 && memlimit_str[0] && safe_uint64(memlimit_str, &memlimit, 10) < 0)
-		lxcfs_error("Failed to convert memory%s.max=%s for cgroup %s",
-			    swap ? ".swap" : "", memlimit_str, cgroup);
 
-	return memlimit;
+	if (ret < 0)
+		return ret;
+
+	if (memlimit_str[0]) {
+		ret = safe_uint64(memlimit_str, &memlimit, 10);
+		if (ret < 0) {
+			lxcfs_error("Failed to convert memory%s.max=%s for cgroup %s",
+				    swap ? ".swap" : "", memlimit_str, cgroup);
+			return ret;
+		}
+	}
+	*limit = memlimit;
+	return 0;
 }
 
 /*
@@ -273,31 +334,44 @@ static char *gnu_dirname(char *path)
 	return path;
 }
 
-static uint64_t get_min_memlimit(const char *cgroup, bool swap)
+/**
+ * Gets a hierarchical memory controller limit, or UINT64_MAX if no limit is
+ * in place. If `swap` is true, reads 'swap' (v2) or 'memsw' (v1); otherwise
+ * reads the memory (RAM) limits.
+ *
+ * @returns 0 on success (and sets `*limit`), < 0 on error
+ */
+static int get_min_memlimit(const char *cgroup, bool swap, uint64_t *limit)
 {
 	__do_free char *copy = NULL;
-	uint64_t memlimit = 0, retlimit = 0;
+	uint64_t memlimit = UINT64_MAX, retlimit = UINT64_MAX;
+	int ret;
 
 	copy = strdup(cgroup);
 	if (!copy)
 		return log_error_errno(0, ENOMEM, "Failed to allocate memory");
 
-	retlimit = get_memlimit(copy, swap);
+	ret = get_memlimit(copy, swap, &retlimit);
+	if (ret < 0)
+		return ret;
 
 	/*
 	 * If the cgroup doesn't start with / (probably won't happen), dirname()
 	 * will terminate with "" instead of "/"
 	 */
-	while (*copy && strcmp(copy, "/") != 0) {
+	while (retlimit != 0 && *copy && strcmp(copy, "/") != 0) {
 		char *it = copy;
 
 		it = gnu_dirname(it);
-		memlimit = get_memlimit(it, swap);
-		if (memlimit > 0 && memlimit < retlimit)
+		ret = get_memlimit(it, swap, &memlimit);
+		if (ret < 0)
+			return ret;
+		if (memlimit < retlimit)
 			retlimit = memlimit;
-	};
+	}
 
-	return retlimit;
+	*limit = retlimit;
+	return 0;
 }
 
 static inline bool startswith(const char *line, const char *pref)
@@ -316,30 +390,30 @@ static void get_swap_info(const char *cgroup, uint64_t memlimit,
 	*swtotal = *swusage = 0;
 	*memswpriority = 1;
 
-	memswlimit = get_min_memlimit(cgroup, true);
-	if (memswlimit > 0) {
-		ret = cgroup_ops->get_memory_swap_current(cgroup_ops, cgroup, &memswusage_str);
-		if (ret < 0 || safe_uint64(memswusage_str, &memswusage, 10) != 0)
-			return;
+	ret = get_min_memlimit(cgroup, true, &memswlimit);
+	if (ret < 0)
+		return;
+	ret = cgroup_ops->get_memory_swap_current(cgroup_ops, cgroup, &memswusage_str);
+	if (ret < 0 || safe_uint64(memswusage_str, &memswusage, 10) < 0)
+		return;
 
-		if (liblxcfs_memory_is_cgroupv2()) {
-			*swtotal = memswlimit / 1024;
-			*swusage = memswusage / 1024;
-		} else {
-			if (memlimit > memswlimit)
-				*swtotal = 0;
-			else
-				*swtotal = (memswlimit - memlimit) / 1024;
-			if (memusage > memswusage || swtotal == 0)
-				*swusage = 0;
-			else
-				*swusage = (memswusage - memusage) / 1024;
-		}
-
-		ret = cgroup_ops->get_memory_swappiness(cgroup_ops, cgroup, &memswpriority_str);
-		if (ret >= 0)
-			safe_uint64(memswpriority_str, memswpriority, 10);
+	if (liblxcfs_memory_is_cgroupv2()) {
+		*swtotal = memswlimit / 1024;
+		*swusage = memswusage / 1024;
+	} else {
+		if (memlimit > memswlimit)
+			*swtotal = 0;
+		else
+			*swtotal = (memswlimit - memlimit) / 1024;
+		if (memusage > memswusage || *swtotal == 0)
+			*swusage = 0;
+		else
+			*swusage = (memswusage - memusage) / 1024;
 	}
+
+	ret = cgroup_ops->get_memory_swappiness(cgroup_ops, cgroup, &memswpriority_str);
+	if (ret >= 0)
+		safe_uint64(memswpriority_str, memswpriority, 10);
 }
 
 static int proc_swaps_read(char *buf, size_t size, off_t offset,
@@ -387,12 +461,12 @@ static int proc_swaps_read(char *buf, size_t size, off_t offset,
 		return read_file_fuse("/proc/swaps", buf, size, d);
 	prune_init_slice(cgroup);
 
-	memlimit = get_min_memlimit(cgroup, false);
-
+	ret = get_min_memlimit(cgroup, false, &memlimit);
+	if (ret < 0)
+		return 0;
 	ret = cgroup_ops->get_memory_current(cgroup_ops, cgroup, &memusage_str);
 	if (ret < 0)
 		return 0;
-
 	if (safe_uint64(memusage_str, &memusage, 10) < 0)
 		lxcfs_error("Failed to convert memusage %s", memusage_str);
 
@@ -414,11 +488,13 @@ static int proc_swaps_read(char *buf, size_t size, off_t offset,
 	}
 
 	if (wants_swap) {
-		/* The total amount of swap is always reported to be the
+		/* For cgroups v1, the total amount of swap is always reported to be the
 		   lesser of the RAM+SWAP limit or the SWAP device size.
 		   This is because the kernel can swap as much as it
 		   wants and not only up to swtotal. */
-		swtotal = memlimit / 1024 + swtotal;
+		if (!liblxcfs_memory_is_cgroupv2())
+			swtotal = memlimit / 1024 + swtotal;
+
 		if (hostswtotal < swtotal) {
 			swtotal = hostswtotal;
 		}
@@ -1273,8 +1349,9 @@ static int proc_meminfo_read(char *buf, size_t size, off_t offset,
 	if (!cgroup_parse_memory_stat(cgroup, &mstat))
 		return read_file_fuse("/proc/meminfo", buf, size, d);
 
-	memlimit = get_min_memlimit(cgroup, false);
-
+	ret = get_min_memlimit(cgroup, false, &memlimit);
+	if (ret < 0)
+		return read_file_fuse("/proc/meminfo", buf, size, d);
 	/*
 	 * Following values are allowed to fail, because swapaccount might be
 	 * turned off for current kernel.
@@ -1306,7 +1383,7 @@ static int proc_meminfo_read(char *buf, size_t size, off_t offset,
 			snprintf(lbuf, 100, "MemFree:        %8" PRIu64 " kB\n", memlimit - memusage);
 			printme = lbuf;
 		} else if (startswith(line, "MemAvailable:")) {
-			snprintf(lbuf, 100, "MemAvailable:   %8" PRIu64 " kB\n", memlimit - memusage + mstat.total_cache / 1024);
+			snprintf(lbuf, 100, "MemAvailable:   %8" PRIu64 " kB\n", memlimit - memusage + (mstat.total_active_file + mstat.total_inactive_file) / 1024);
 			printme = lbuf;
 		} else if (startswith(line, "SwapTotal:")) {
 			if (wants_swap) {
@@ -1314,11 +1391,10 @@ static int proc_meminfo_read(char *buf, size_t size, off_t offset,
 
 				sscanf(line + STRLITERALLEN("SwapTotal:"), "%" PRIu64, &hostswtotal);
 
-				/* The total amount of swap is always reported to be the
+				/* In cgroups v1, the total amount of swap is always reported to be the
 				   lesser of the RAM+SWAP limit or the SWAP device size.
 				   This is because the kernel can swap as much as it
 				   wants and not only up to swtotal. */
-
 				if (!liblxcfs_memory_is_cgroupv2())
 					swtotal += memlimit;
 
